signature cnfTheory =
sig
  type thm = Thm.thm
  
  (*  Definitions  *)
    val badNtRules_def : thm
    val badTmnlRules_def : thm
    val cnf_def : thm
    val elimNT_curried_def : thm
    val elimNT_tupled_primitive_def : thm
    val elim_curried_def : thm
    val elim_tupled_primitive_def : thm
    val eqLang_def : thm
    val noNewNts_def : thm
    val ruleNonTmnls_primitive_def : thm
    val ruleTmnls_primitive_def : thm
    val trans1Tmnl_def : thm
    val trans2NT_def : thm
  
  (*  Theorems  *)
    val cnf_lemma : thm
    val cnf_lemma1 : thm
    val cnf_lemma2 : thm
    val cnf_lemma2NT : thm
    val elimNT_def : thm
    val elimNT_ind : thm
    val elim_def : thm
    val elim_ind : thm
    val lemma14_b : thm
    val lemma9_a : thm
    val lemma9_b : thm
    val ruleNonTmnls_def : thm
    val ruleNonTmnls_ind : thm
    val ruleTmnlsAlt : thm
    val ruleTmnls_def : thm
    val ruleTmnls_ind : thm
    val thm4_5 : thm
  
  val cnf_grammars : type_grammar.grammar * term_grammar.grammar
  
  
(*
   [grammarDef] Parent theory of "cnf"
   
   [badNtRules_def]  Definition
      
      |- ∀g. badNtRules g = SUM (MAP ruleNonTmnls (rules g))
   
   [badTmnlRules_def]  Definition
      
      |- ∀g. badTmnlRules g = SUM (MAP ruleTmnls (rules g))
   
   [cnf_def]  Definition
      
      |- ∀g. cnf g ⇔ (badNtRules g = 0) ∧ (badTmnlRules g = 0)
   
   [elimNT_curried_def]  Definition
      
      |- ∀x x1. elimNT x x1 = elimNT_tupled (x,x1)
   
   [elimNT_tupled_primitive_def]  Definition
      
      |- elimNT_tupled =
         WFREC
           (@R.
              WF R ∧
              (∀nt2 nt1 sf nt s.
                 s ≠ NTS nt ⇒ R (sf,rule nt [nt1; nt2]) (s::sf,rule nt [nt1; nt2])) ∧
              ∀nt2 nt1 sf nt s.
                (s = NTS nt) ⇒ R (sf,rule nt [nt1; nt2]) (s::sf,rule nt [nt1; nt2]))
           (λelimNT_tupled a.
              case a of
                 ([],rule nt []) -> ARB
              || ([],rule nt [nt1]) -> ARB
              || ([],rule nt [nt1; nt2]) -> I []
              || ([],rule nt (nt1::nt2::v10::v11)) -> ARB
              || (s::sf,rule nt' []) -> ARB
              || (s::sf,rule nt' [nt1']) -> ARB
              || (s::sf,rule nt' [nt1'; nt2']) ->
                   I
                     (if s = NTS nt' then
                        [nt1'; nt2'] ++ elimNT_tupled (sf,rule nt' [nt1'; nt2'])
                      else
                        s::elimNT_tupled (sf,rule nt' [nt1'; nt2']))
              || (s::sf,rule nt' (nt1'::nt2'::v20::v21)) -> ARB)
   
   [elim_curried_def]  Definition
      
      |- ∀x x1. elim x x1 = elim_tupled (x,x1)
   
   [elim_tupled_primitive_def]  Definition
      
      |- elim_tupled =
         WFREC (@R. WF R)
           (λelim_tupled a.
              case a of
                 (sf,rule nt []) -> ARB
              || (sf,rule nt [t]) -> I (MAP (λe. if e = NTS nt then t else e) sf)
              || (sf,rule nt (t::v6::v7)) -> ARB)
   
   [eqLang_def]  Definition
      
      |- ∀g g'. eqLang g g' ⇔ (language g = language g')
   
   [noNewNts_def]  Definition
      
      |- ∀g sf. noNewNts g sf ⇔ ∀e. MEM e sf ⇒ isNonTmnlSym e ⇒ e ∈ nonTerminals g
   
   [ruleNonTmnls_primitive_def]  Definition
      
      |- ruleNonTmnls =
         WFREC (@R. WF R)
           (λruleNonTmnls a.
              case a of
                 rule l [] -> I 0
              || rule l [t] -> I 0
              || rule l [t; t2] -> I 0
              || rule l (t::t2::v12::v13) ->
                   I (LENGTH (FILTER isNonTmnlSym (t::t2::v12::v13))))
   
   [ruleTmnls_primitive_def]  Definition
      
      |- ruleTmnls =
         WFREC (@R. WF R)
           (λruleTmnls a.
              case a of
                 rule l [] -> I (LENGTH (FILTER isTmnlSym []))
              || rule l [t] -> I 0
              || rule l (t::v8::v9) -> I (LENGTH (FILTER isTmnlSym (t::v8::v9))))
   
   [trans1Tmnl_def]  Definition
      
      |- ∀nt t g g'.
           trans1Tmnl nt t g g' ⇔
           ∃l r p s.
             MEM (rule l r) (rules g) ∧ (r = p ++ [t] ++ s) ∧ (p ≠ [] ∨ s ≠ []) ∧
             isTmnlSym t ∧ NTS nt ∉ nonTerminals g ∧
             (rules g' =
              delete (rule l r) (rules g) ++
              [rule nt [t]; rule l (p ++ [NTS nt] ++ s)]) ∧
             (startSym g' = startSym g)
   
   [trans2NT_def]  Definition
      
      |- ∀nt nt1 nt2 g g'.
           trans2NT nt nt1 nt2 g g' ⇔
           ∃l r p s.
             MEM (rule l r) (rules g) ∧ (r = p ++ [nt1; nt2] ++ s) ∧
             (p ≠ [] ∨ s ≠ []) ∧ isNonTmnlSym nt1 ∧ isNonTmnlSym nt2 ∧
             NTS nt ∉ nonTerminals g ∧
             (rules g' =
              delete (rule l r) (rules g) ++
              [rule nt [nt1; nt2]; rule l (p ++ [NTS nt] ++ s)]) ∧
             (startSym g' = startSym g)
   
   [cnf_lemma]  Theorem
      
      |- ∀g g' g''.
           (λx y. ∃nt t. trans1Tmnl nt t x y)^* g g' ⇒
           (λx y. ∃nt nt1 nt2. trans2NT nt nt1 nt2 x y)^* g' g'' ⇒
           eqLang g g''
   
   [cnf_lemma1]  Theorem
      
      |- ∀g g' nt t. trans1Tmnl nt t g g' ⇒ (language g = language g')
   
   [cnf_lemma2]  Theorem
      
      |- ∀g g'. (λx y. ∃nt t. trans1Tmnl nt t x y)^* g g' ⇒ eqLang g g'
   
   [cnf_lemma2NT]  Theorem
      
      |- ∀g g'. (λx y. ∃nt nt1 nt2. trans2NT nt nt1 nt2 x y)^* g g' ⇒ eqLang g g'
   
   [elimNT_def]  Theorem
      
      |- (∀nt2 nt1 nt. elimNT [] (rule nt [nt1; nt2]) = []) ∧
         ∀sf s nt2 nt1 nt.
           elimNT (s::sf) (rule nt [nt1; nt2]) =
           if s = NTS nt then
             [nt1; nt2] ++ elimNT sf (rule nt [nt1; nt2])
           else
             s::elimNT sf (rule nt [nt1; nt2])
   
   [elimNT_ind]  Theorem
      
      |- ∀P.
           (∀nt nt1 nt2. P [] (rule nt [nt1; nt2])) ∧
           (∀s sf nt nt1 nt2.
              (s ≠ NTS nt ⇒ P sf (rule nt [nt1; nt2])) ∧
              ((s = NTS nt) ⇒ P sf (rule nt [nt1; nt2])) ⇒
              P (s::sf) (rule nt [nt1; nt2])) ∧ (∀nt. P [] (rule nt [])) ∧
           (∀nt nt1. P [] (rule nt [nt1])) ∧
           (∀nt nt1 nt2 v12 v13. P [] (rule nt (nt1::nt2::v12::v13))) ∧
           (∀s sf nt. P (s::sf) (rule nt [])) ∧
           (∀s sf nt nt1. P (s::sf) (rule nt [nt1])) ∧
           (∀s sf nt nt1 nt2 v22 v23. P (s::sf) (rule nt (nt1::nt2::v22::v23))) ⇒
           ∀v v1. P v v1
   
   [elim_def]  Theorem
      
      |- elim sf (rule nt [t]) = MAP (λe. if e = NTS nt then t else e) sf
   
   [elim_ind]  Theorem
      
      |- ∀P.
           (∀sf nt t. P sf (rule nt [t])) ∧ (∀sf nt. P sf (rule nt [])) ∧
           (∀sf nt t v8 v9. P sf (rule nt (t::v8::v9))) ⇒
           ∀v v1. P v v1
   
   [lemma14_b]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] []
      |- ∀g g'.
           (λx y. ∃nt nt1 nt2. trans2NT nt nt1 nt2 x y)^* g g' ⇒
           (badTmnlRules g = 0) ⇒
           (badTmnlRules g' = 0)
   
   [lemma9_a]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] []
      |- INFINITE UNIV ⇒
         ∃g'. (λx y. ∃nt t. trans1Tmnl nt t x y)^* g g' ∧ (badTmnlRules g' = 0)
   
   [lemma9_b]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] []
      |- INFINITE UNIV ⇒
         (badTmnlRules g = 0) ⇒
         ∃g'.
           (λx y. ∃nt nt1 nt2. trans2NT nt nt1 nt2 x y)^* g g' ∧ (badNtRules g' = 0)
   
   [ruleNonTmnls_def]  Theorem
      
      |- (ruleNonTmnls (rule l []) = 0) ∧ (ruleNonTmnls (rule l [t]) = 0) ∧
         (ruleNonTmnls (rule l [t1; t2]) = 0) ∧
         (ruleNonTmnls (rule l (v2::v6::v10::v11)) =
          LENGTH (FILTER isNonTmnlSym (v2::v6::v10::v11)))
   
   [ruleNonTmnls_ind]  Theorem
      
      |- ∀P.
           (∀l. P (rule l [])) ∧ (∀l t. P (rule l [t])) ∧
           (∀l t1 t2. P (rule l [t1; t2])) ∧
           (∀l v2 v6 v10 v11. P (rule l (v2::v6::v10::v11))) ⇒
           ∀v. P v
   
   [ruleTmnlsAlt]  Theorem
      
      |- ∀l r.
           ruleTmnls (rule l r) =
           if LENGTH r ≤ 1 then 0 else LENGTH (FILTER isTmnlSym r)
   
   [ruleTmnls_def]  Theorem
      
      |- (ruleTmnls (rule l [t]) = 0) ∧
         (ruleTmnls (rule l []) = LENGTH (FILTER isTmnlSym [])) ∧
         (ruleTmnls (rule l (v2::v6::v7)) = LENGTH (FILTER isTmnlSym (v2::v6::v7)))
   
   [ruleTmnls_ind]  Theorem
      
      |- ∀P.
           (∀l t. P (rule l [t])) ∧ (∀l. P (rule l [])) ∧
           (∀l v2 v6 v7. P (rule l (v2::v6::v7))) ⇒
           ∀v. P v
   
   [thm4_5]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] []
      |- ∀g. INFINITE UNIV ⇒ ∃g'. cnf g' ∧ eqLang g g'
   
   
*)
end
