signature eProdsTheory =
sig
  type thm = Thm.thm
  
  (*  Definitions  *)
    val munge0_def : thm
    val munge_def : thm
    val negr_def : thm
    val no_rhs_def : thm
    val no_rule_curried_def : thm
    val no_rule_tupled_primitive_def : thm
    val nullable_def : thm
  
  (*  Theorems  *)
    val eq_snegr : thm
    val negr_exists : thm
    val negr_r15 : thm
    val no_rule_def : thm
    val no_rule_ind : thm
    val thm4_3 : thm
  
  val eProds_grammars : type_grammar.grammar * term_grammar.grammar
  
  
(*
   [grammarDef] Parent theory of "eProds"
   
   [munge0_def]  Definition
      
      |- (∀g. munge0 g [] = [[]]) ∧
         ∀g s sl.
           munge0 g (s::sl) =
           if nullable g [s] then
             MAP (CONS s) (munge0 g sl) ++ munge0 g sl
           else
             MAP (CONS s) (munge0 g sl)
   
   [munge_def]  Definition
      
      |- ∀g p.
           munge g p =
           {rule l r' | ∃r. MEM (rule l r) p ∧ MEM r' (munge0 g r) ∧ r' ≠ []}
   
   [negr_def]  Definition
      
      |- ∀g g'.
           negr g g' ⇔
           (∀e. MEM e (rules g') ⇔ e ∈ {rule l r | rule l r ∈ munge g (rules g)}) ∧
           (startSym g' = startSym g)
   
   [no_rhs_def]  Definition
      
      |- ∀g l1 l2. no_rhs g l1 l2 ⇔ MEM l2 (munge0 g l1)
   
   [no_rule_curried_def]  Definition
      
      |- ∀x x1 x2. no_rule x x1 x2 ⇔ no_rule_tupled (x,x1,x2)
   
   [no_rule_tupled_primitive_def]  Definition
      
      |- no_rule_tupled =
         WFREC (@R. WF R)
           (λno_rule_tupled a.
              case a of
                 (g,rule l1 r1,rule l2 r2) -> I ((l1 = l2) ∧ MEM r2 (munge0 g r1)))
   
   [nullable_def]  Definition
      
      |- ∀g sl. nullable g sl ⇔ (derives g)^* sl []
   
   [eq_snegr]  Theorem
      
      |- negr g g' ⇒ (startSym g = startSym g')
   
   [negr_exists]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] [] |- ∀g. ∃g'. negr g g'
   
   [negr_r15]  Theorem
      
      |- ∀s sf.
           (derives g)^* s sf ⇒
           negr g g' ⇒
           (s = [NTS (startSym g)]) ⇒
           ∀sf'. no_rhs g sf sf' ⇒ sf' ≠ [] ⇒ (derives g')^* s sf'
   
   [no_rule_def]  Theorem
      
      |- no_rule g (rule l1 r1) (rule l2 r2) ⇔ (l1 = l2) ∧ MEM r2 (munge0 g r1)
   
   [no_rule_ind]  Theorem
      
      |- ∀P. (∀g l1 r1 l2 r2. P g (rule l1 r1) (rule l2 r2)) ⇒ ∀v v1 v2. P v v1 v2
   
   [thm4_3]  Theorem
      
      |- negr g g' ⇒ [] ∉ language g ⇒ (language g = language g')
   
   
*)
end
