signature parseTreeTheory =
sig
  type thm = Thm.thm
  
  (*  Definitions  *)
    val checkRules_def : thm
    val getSymbols_primitive_def : thm
    val isLeaf_primitive_def : thm
    val isNode_primitive_def : thm
    val leaves_def : thm
    val ptheight_def : thm
    val ptree2Sym_primitive_def : thm
    val ptreeNodeSym_primitive_def : thm
    val ptreeSubtSyms_primitive_def : thm
    val ptreeSubtree_primitive_def : thm
    val ptreeToRules_def : thm
    val ptree_TY_DEF : thm
    val ptree_case_def : thm
    val ptree_repfns : thm
    val ptree_size_def : thm
    val ptsize_def : thm
    val validptree_defn_curried_def : thm
    val validptree_defn_tupled_primitive_def : thm
  
  (*  Theorems  *)
    val checkRules_append : thm
    val datatype_ptree : thm
    val flat_leaves : thm
    val getSymbols_def : thm
    val getSymbols_ind : thm
    val getSymsEqptree2Sym : thm
    val getSyms_append : thm
    val getSyms_len : thm
    val getSyms_nil : thm
    val getSyms_rev : thm
    val isLeaf_def : thm
    val isLeaf_ind : thm
    val isNode_def : thm
    val isNode_ind : thm
    val mapSymsptree : thm
    val ptree2Sym_def : thm
    val ptree2Sym_ind : thm
    val ptreeNodeSym_def : thm
    val ptreeNodeSym_ind : thm
    val ptreeSubtSyms_def : thm
    val ptreeSubtSyms_ind : thm
    val ptreeSubtree_def : thm
    val ptreeSubtree_ind : thm
    val ptreeToRules_append : thm
    val ptree_11 : thm
    val ptree_Axiom : thm
    val ptree_case_cong : thm
    val ptree_distinct : thm
    val ptree_induction : thm
    val ptree_nchotomy : thm
    val ptsizel_append : thm
    val take1_getSyms : thm
    val take_getSyms : thm
    val validptree : thm
    val validptree_ind : thm
  
  val parseTree_grammars : type_grammar.grammar * term_grammar.grammar
  
  val parseTree_rwts : simpLib.ssfrag
(*
   [grammarDef] Parent theory of "parseTree"
   
   [checkRules_def]  Definition
      
      |- (∀rls. checkRules [] rls ⇔ T) ∧
         ∀h t rls. checkRules (h::t) rls ⇔ MEM h rls ∧ checkRules t rls
   
   [getSymbols_primitive_def]  Definition
      
      |- getSymbols =
         WFREC
           (@R.
              WF R ∧ (∀tmnl t. R t (Leaf tmnl::t)) ∧
              ∀ptlist nt t. R t (Node nt ptlist::t))
           (λgetSymbols a.
              case a of
                 [] -> I []
              || Leaf tmnl::t -> I (TS tmnl::getSymbols t)
              || Node nt ptlist::t -> I (NTS nt::getSymbols t))
   
   [isLeaf_primitive_def]  Definition
      
      |- isLeaf =
         WFREC (@R. WF R) (λisLeaf a. case a of Leaf v2 -> I T || Node v0 v1 -> I F)
   
   [isNode_primitive_def]  Definition
      
      |- isNode =
         WFREC (@R. WF R) (λisNode a. case a of Leaf v2 -> I F || Node v0 v1 -> I T)
   
   [leaves_def]  Definition
      
      |- (∀tm. leaves (Leaf tm) = [tm]) ∧
         (∀nt ptlist. leaves (Node nt ptlist) = cleaves ptlist) ∧ (cleaves [] = []) ∧
         ∀h t. cleaves (h::t) = leaves h ++ cleaves t
   
   [ptheight_def]  Definition
      
      |- (∀tmnl. ptheight (Leaf tmnl) = 0) ∧
         (∀nt ptlist. ptheight (Node nt ptlist) = 1 + ptheightl ptlist) ∧
         (ptheightl [] = 0) ∧ ∀h t. ptheightl (h::t) = ptheight h + ptheightl t
   
   [ptree2Sym_primitive_def]  Definition
      
      |- ptree2Sym =
         WFREC (@R. WF R)
           (λptree2Sym a.
              case a of Leaf tm -> I (TS tm) || Node nt ptl -> I (NTS nt))
   
   [ptreeNodeSym_primitive_def]  Definition
      
      |- ptreeNodeSym =
         WFREC (@R. WF R)
           (λptreeNodeSym a.
              case a of Leaf tm -> I (TS tm) || Node nt tl -> I (NTS nt))
   
   [ptreeSubtSyms_primitive_def]  Definition
      
      |- ptreeSubtSyms =
         WFREC (@R. WF R)
           (λptreeSubtSyms a.
              case a of Leaf tm -> I [] || Node nt tl -> I (MAP ptreeNodeSym tl))
   
   [ptreeSubtree_primitive_def]  Definition
      
      |- ptreeSubtree =
         WFREC (@R. WF R)
           (λptreeSubtree a. case a of Leaf tm -> I [] || Node x l -> I l)
   
   [ptreeToRules_def]  Definition
      
      |- (∀tm. ptreeToRules (Leaf tm) = []) ∧
         (∀nt ptl.
            ptreeToRules (Node nt ptl) =
            rule nt (getSymbols ptl)::ptreeToRules2 ptl) ∧ (ptreeToRules2 [] = []) ∧
         ∀h t. ptreeToRules2 (h::t) = ptreeToRules h ++ ptreeToRules2 t
   
   [ptree_TY_DEF]  Definition
      
      |- ∃rep.
           TYPE_DEFINITION
             (λa0'.
                ∀'ptree' 'list @ind_typeparseTree0' .
                  (∀a0'.
                     (∃a.
                        a0' =
                        (λa. ind_type$CONSTR 0 (a,ARB) (λn. ind_type$BOTTOM)) a) ∨
                     (∃a0 a1.
                        (a0' =
                         (λa0 a1.
                            ind_type$CONSTR (SUC 0) (ARB,a0)
                              (ind_type$FCONS a1 (λn. ind_type$BOTTOM))) a0 a1) ∧
                        'list @ind_typeparseTree0' a1) ⇒
                     'ptree' a0') ∧
                  (∀a1'.
                     (a1' =
                      ind_type$CONSTR (SUC (SUC 0)) (ARB,ARB)
                        (λn. ind_type$BOTTOM)) ∨
                     (∃a0 a1.
                        (a1' =
                         (λa0 a1.
                            ind_type$CONSTR (SUC (SUC (SUC 0))) (ARB,ARB)
                              (ind_type$FCONS a0
                                 (ind_type$FCONS a1 (λn. ind_type$BOTTOM)))) a0 a1) ∧
                        'ptree' a0 ∧ 'list @ind_typeparseTree0' a1) ⇒
                     'list @ind_typeparseTree0' a1') ⇒
                  'ptree' a0') rep
   
   [ptree_case_def]  Definition
      
      |- (∀f f1 a. ptree_case f f1 (Leaf a) = f a) ∧
         ∀f f1 a0 a1. ptree_case f f1 (Node a0 a1) = f1 a0 a1
   
   [ptree_repfns]  Definition
      
      |- (∀a. mk_ptree (dest_ptree a) = a) ∧
         ∀r.
           (λa0'.
              ∀'ptree' 'list @ind_typeparseTree0' .
                (∀a0'.
                   (∃a.
                      a0' =
                      (λa. ind_type$CONSTR 0 (a,ARB) (λn. ind_type$BOTTOM)) a) ∨
                   (∃a0 a1.
                      (a0' =
                       (λa0 a1.
                          ind_type$CONSTR (SUC 0) (ARB,a0)
                            (ind_type$FCONS a1 (λn. ind_type$BOTTOM))) a0 a1) ∧
                      'list @ind_typeparseTree0' a1) ⇒
                   'ptree' a0') ∧
                (∀a1'.
                   (a1' =
                    ind_type$CONSTR (SUC (SUC 0)) (ARB,ARB) (λn. ind_type$BOTTOM)) ∨
                   (∃a0 a1.
                      (a1' =
                       (λa0 a1.
                          ind_type$CONSTR (SUC (SUC (SUC 0))) (ARB,ARB)
                            (ind_type$FCONS a0
                               (ind_type$FCONS a1 (λn. ind_type$BOTTOM)))) a0 a1) ∧
                      'ptree' a0 ∧ 'list @ind_typeparseTree0' a1) ⇒
                   'list @ind_typeparseTree0' a1') ⇒
                'ptree' a0') r ⇔ (dest_ptree (mk_ptree r) = r)
   
   [ptree_size_def]  Definition
      
      |- (∀f f1 a. ptree_size f f1 (Leaf a) = 1 + f1 a) ∧
         (∀f f1 a0 a1.
            ptree_size f f1 (Node a0 a1) = 1 + (f a0 + ptree1_size f f1 a1)) ∧
         (∀f f1. ptree1_size f f1 [] = 0) ∧
         ∀f f1 a0 a1.
           ptree1_size f f1 (a0::a1) = 1 + (ptree_size f f1 a0 + ptree1_size f f1 a1)
   
   [ptsize_def]  Definition
      
      |- (∀tmnl. ptsize (Leaf tmnl) = 1) ∧
         (∀nt ptlist. ptsize (Node nt ptlist) = 1 + ptsizel ptlist) ∧
         (ptsizel [] = 0) ∧ ∀h t. ptsizel (h::t) = ptsize h + ptsizel t
   
   [validptree_defn_curried_def]  Definition
      
      |- ∀x x1. validptree x x1 ⇔ validptree_defn_tupled (x,x1)
   
   [validptree_defn_tupled_primitive_def]  Definition
      
      |- validptree_defn_tupled =
         WFREC (@R. WF R ∧ ∀n g ptl e. MEM e ptl ∧ isNode e ⇒ R (g,e) (g,Node n ptl))
           (λvalidptree_defn_tupled a.
              case a of
                 (g,Leaf tm) -> I F
              || (g,Node n ptl) ->
                   I
                     (MEM (rule n (getSymbols ptl)) (rules g) ∧
                      ∀e. MEM e ptl ⇒ isNode e ⇒ validptree_defn_tupled (g,e)))
   
   [checkRules_append]  Theorem
      
      |- checkRules (l1 ++ l2) rs ⇔ checkRules l1 rs ∧ checkRules l2 rs
   
   [datatype_ptree]  Theorem
      
      |- DATATYPE (ptree Leaf Node)
   
   [flat_leaves]  Theorem
      
      |- ∀l. leaves (Node n l) = FLAT (MAP leaves l)
   
   [getSymbols_def]  Theorem
      
      |- (getSymbols [] = []) ∧
         (∀tmnl t. getSymbols (Leaf tmnl::t) = TS tmnl::getSymbols t) ∧
         ∀t ptlist nt. getSymbols (Node nt ptlist::t) = NTS nt::getSymbols t
   
   [getSymbols_ind]  Theorem
      
      |- ∀P.
           P [] ∧ (∀tmnl t. P t ⇒ P (Leaf tmnl::t)) ∧
           (∀nt ptlist t. P t ⇒ P (Node nt ptlist::t)) ⇒
           ∀v. P v
   
   [getSymsEqptree2Sym]  Theorem
      
      |- getSymbols l = MAP ptree2Sym l
   
   [getSyms_append]  Theorem
      
      |- getSymbols (l1 ++ l2) = getSymbols l1 ++ getSymbols l2
   
   [getSyms_len]  Theorem
      
      |- ∀l. LENGTH l = LENGTH (getSymbols l)
   
   [getSyms_nil]  Theorem
      
      |- (getSymbols l = []) ⇒ (l = [])
   
   [getSyms_rev]  Theorem
      
      |- getSymbols (REVERSE x) = REVERSE (getSymbols x)
   
   [isLeaf_def]  Theorem
      
      |- (isLeaf (Node v0 v1) ⇔ F) ∧ (isLeaf (Leaf v2) ⇔ T)
   
   [isLeaf_ind]  Theorem
      
      |- ∀P. (∀v0 v1. P (Node v0 v1)) ∧ (∀v2. P (Leaf v2)) ⇒ ∀v. P v
   
   [isNode_def]  Theorem
      
      |- (isNode (Node v0 v1) ⇔ T) ∧ (isNode (Leaf v2) ⇔ F)
   
   [isNode_ind]  Theorem
      
      |- ∀P. (∀v0 v1. P (Node v0 v1)) ∧ (∀v2. P (Leaf v2)) ⇒ ∀v. P v
   
   [mapSymsptree]  Theorem
      
      |- MAP (ptree2Sym o SND) p = getSymbols (MAP SND p)
   
   [ptree2Sym_def]  Theorem
      
      |- (ptree2Sym (Node nt ptl) = NTS nt) ∧ (ptree2Sym (Leaf tm) = TS tm)
   
   [ptree2Sym_ind]  Theorem
      
      |- ∀P. (∀nt ptl. P (Node nt ptl)) ∧ (∀tm. P (Leaf tm)) ⇒ ∀v. P v
   
   [ptreeNodeSym_def]  Theorem
      
      |- (ptreeNodeSym (Node nt tl) = NTS nt) ∧ (ptreeNodeSym (Leaf tm) = TS tm)
   
   [ptreeNodeSym_ind]  Theorem
      
      |- ∀P. (∀nt tl. P (Node nt tl)) ∧ (∀tm. P (Leaf tm)) ⇒ ∀v. P v
   
   [ptreeSubtSyms_def]  Theorem
      
      |- (ptreeSubtSyms (Node nt tl) = MAP ptreeNodeSym tl) ∧
         (ptreeSubtSyms (Leaf tm) = [])
   
   [ptreeSubtSyms_ind]  Theorem
      
      |- ∀P. (∀nt tl. P (Node nt tl)) ∧ (∀tm. P (Leaf tm)) ⇒ ∀v. P v
   
   [ptreeSubtree_def]  Theorem
      
      |- (ptreeSubtree (Node x l) = l) ∧ (ptreeSubtree (Leaf tm) = [])
   
   [ptreeSubtree_ind]  Theorem
      
      |- ∀P. (∀x l. P (Node x l)) ∧ (∀tm. P (Leaf tm)) ⇒ ∀v. P v
   
   [ptreeToRules_append]  Theorem
      
      |- ptreeToRules2 (l1 ++ l2) = ptreeToRules2 l1 ++ ptreeToRules2 l2
   
   [ptree_11]  Theorem
      
      |- (∀a a'. (Leaf a = Leaf a') ⇔ (a = a')) ∧
         ∀a0 a1 a0' a1'. (Node a0 a1 = Node a0' a1') ⇔ (a0 = a0') ∧ (a1 = a1')
   
   [ptree_Axiom]  Theorem
      
      |- ∀f0 f1 f2 f3.
           ∃fn0 fn1.
             (∀a. fn0 (Leaf a) = f0 a) ∧
             (∀a0 a1. fn0 (Node a0 a1) = f1 a0 a1 (fn1 a1)) ∧ (fn1 [] = f2) ∧
             ∀a0 a1. fn1 (a0::a1) = f3 a0 a1 (fn0 a0) (fn1 a1)
   
   [ptree_case_cong]  Theorem
      
      |- ∀M M' f f1.
           (M = M') ∧ (∀a. (M' = Leaf a) ⇒ (f a = f' a)) ∧
           (∀a0 a1. (M' = Node a0 a1) ⇒ (f1 a0 a1 = f1' a0 a1)) ⇒
           (ptree_case f f1 M = ptree_case f' f1' M')
   
   [ptree_distinct]  Theorem
      
      |- ∀a1 a0 a. Leaf a ≠ Node a0 a1
   
   [ptree_induction]  Theorem
      
      |- ∀P0 P1.
           (∀t. P0 (Leaf t)) ∧ (∀l. P1 l ⇒ ∀n. P0 (Node n l)) ∧ P1 [] ∧
           (∀p l. P0 p ∧ P1 l ⇒ P1 (p::l)) ⇒
           (∀p. P0 p) ∧ ∀l. P1 l
   
   [ptree_nchotomy]  Theorem
      
      |- ∀pp. (∃t. pp = Leaf t) ∨ ∃n l. pp = Node n l
   
   [ptsizel_append]  Theorem
      
      |- ptsizel (l1 ++ l2) = ptsizel l1 + ptsizel l2
   
   [take1_getSyms]  Theorem
      
      |- ∀l n. LENGTH l ≥ n ⇒ (take1 n (getSymbols l) = getSymbols (take1 n l))
   
   [take_getSyms]  Theorem
      
      |- ∀n l.
           (take n (MAP SND l) = SOME x) ⇒
           (THE (take n (getSymbols (MAP SND l))) =
            getSymbols (THE (take n (MAP SND l))))
   
   [validptree]  Theorem
      
      |- (validptree g (Node n ptl) ⇔
          MEM (rule n (getSymbols ptl)) (rules g) ∧
          ∀e. MEM e ptl ⇒ isNode e ⇒ validptree g e) ∧ (validptree g (Leaf tm) ⇔ F)
   
   [validptree_ind]  Theorem
      
      |- ∀P.
           (∀g n ptl. (∀e. MEM e ptl ∧ isNode e ⇒ P g e) ⇒ P g (Node n ptl)) ∧
           (∀g tm. P g (Leaf tm)) ⇒
           ∀v v1. P v v1
   
   
*)
end
