signature regexpTheory =
sig
  type thm = Thm.thm
  
  (*  Definitions  *)
    val conc_def : thm
    val distinctldNts_def : thm
    val isNonTmnlSym_def : thm
    val isTmnlSym_def : thm
    val lang_def : thm
    val ldNts_def : thm
    val nonTmnlSym_def : thm
    val nts2str_def : thm
    val rexp_TY_DEF : thm
    val rexp_case_def : thm
    val rexp_repfns : thm
    val rexp_size_def : thm
    val star : thm
    val starn_curried_def : thm
    val starn_tupled_primitive_def : thm
    val symToStr_def : thm
    val symbol_TY_DEF : thm
    val symbol_case_def : thm
    val symbol_repfns : thm
    val symbol_size_def : thm
    val tmnlSym_def : thm
    val toTmnlSym_def : thm
    val ts2str_def : thm
    val union_def : thm
  
  (*  Theorems  *)
    val datatype_rexp : thm
    val datatype_symbol : thm
    val ldNtsApp : thm
    val listCompLens : thm
    val memdistNtsApp : thm
    val memdistNtsApp' : thm
    val memldNts : thm
    val rexp_11 : thm
    val rexp_Axiom : thm
    val rexp_case_cong : thm
    val rexp_distinct : thm
    val rexp_induction : thm
    val rexp_nchotomy : thm
    val rightnt : thm
    val star_cases : thm
    val star_ind : thm
    val star_rules : thm
    val starn_def : thm
    val starn_ind : thm
    val symListDiv : thm
    val sym_r1a : thm
    val sym_r1b : thm
    val sym_r2 : thm
    val sym_r3 : thm
    val sym_r3b : thm
    val sym_r4 : thm
    val sym_r4b : thm
    val sym_r5 : thm
    val sym_r6 : thm
    val sym_r7 : thm
    val sym_r7b : thm
    val symbol_11 : thm
    val symbol_Axiom : thm
    val symbol_case_cong : thm
    val symbol_distinct : thm
    val symbol_induction : thm
    val symbol_nchotomy : thm
    val symlistBdwn : thm
    val symlistdiv3 : thm
    val symlistnil : thm
  
  val regexp_grammars : type_grammar.grammar * term_grammar.grammar
  
  
(*
   [listLemmas] Parent theory of "regexp"
   
   [conc_def]  Definition
      
      |- ∀as bs. conc as bs = {s | ∃u v. u ∈ as ∧ v ∈ bs ∧ (s = u ++ v)}
   
   [distinctldNts_def]  Definition
      
      |- ∀dl. distinctldNts dl = rmDupes (ldNts dl)
   
   [isNonTmnlSym_def]  Definition
      
      |- ∀sym. isNonTmnlSym sym ⇔ (∃s. sym = NTS s) ∨ F
   
   [isTmnlSym_def]  Definition
      
      |- ∀sym. isTmnlSym sym ⇔ ∃s. sym = TS s
   
   [lang_def]  Definition
      
      |- (lang EmptyLang = {}) ∧ (∀tmnl. lang (Atom tmnl) = {[symToStr tmnl]}) ∧
         (∀r s. lang (Union r s) = lang r ∪ lang s) ∧
         (∀r s. lang (Conc r s) = conc (lang r) (lang s)) ∧
         ∀r. lang (Star r) = star (lang r)
   
   [ldNts_def]  Definition
      
      |- ∀dl. ldNts dl = FILTER isNonTmnlSym (FLAT dl)
   
   [nonTmnlSym_def]  Definition
      
      |- ∀ntmnl. nonTmnlSym (NTS ntmnl) = ntmnl
   
   [nts2str_def]  Definition
      
      |- ∀s. nts2str (NTS s) = s
   
   [rexp_TY_DEF]  Definition
      
      |- ∃rep.
           TYPE_DEFINITION
             (λa0'.
                ∀'rexp' .
                  (∀a0'.
                     (a0' = ind_type$CONSTR 0 ARB (λn. ind_type$BOTTOM)) ∨
                     (∃a.
                        a0' =
                        (λa. ind_type$CONSTR (SUC 0) a (λn. ind_type$BOTTOM)) a) ∨
                     (∃a0 a1.
                        (a0' =
                         (λa0 a1.
                            ind_type$CONSTR (SUC (SUC 0)) ARB
                              (ind_type$FCONS a0
                                 (ind_type$FCONS a1 (λn. ind_type$BOTTOM)))) a0 a1) ∧
                        'rexp' a0 ∧ 'rexp' a1) ∨
                     (∃a0 a1.
                        (a0' =
                         (λa0 a1.
                            ind_type$CONSTR (SUC (SUC (SUC 0))) ARB
                              (ind_type$FCONS a0
                                 (ind_type$FCONS a1 (λn. ind_type$BOTTOM)))) a0 a1) ∧
                        'rexp' a0 ∧ 'rexp' a1) ∨
                     (∃a.
                        (a0' =
                         (λa.
                            ind_type$CONSTR (SUC (SUC (SUC (SUC 0)))) ARB
                              (ind_type$FCONS a (λn. ind_type$BOTTOM))) a) ∧
                        'rexp' a) ⇒
                     'rexp' a0') ⇒
                  'rexp' a0') rep
   
   [rexp_case_def]  Definition
      
      |- (∀v f f1 f2 f3. rexp_case v f f1 f2 f3 EmptyLang = v) ∧
         (∀v f f1 f2 f3 a. rexp_case v f f1 f2 f3 (Atom a) = f a) ∧
         (∀v f f1 f2 f3 a0 a1. rexp_case v f f1 f2 f3 (Union a0 a1) = f1 a0 a1) ∧
         (∀v f f1 f2 f3 a0 a1. rexp_case v f f1 f2 f3 (Conc a0 a1) = f2 a0 a1) ∧
         ∀v f f1 f2 f3 a. rexp_case v f f1 f2 f3 (Star a) = f3 a
   
   [rexp_repfns]  Definition
      
      |- (∀a. mk_rexp (dest_rexp a) = a) ∧
         ∀r.
           (λa0'.
              ∀'rexp' .
                (∀a0'.
                   (a0' = ind_type$CONSTR 0 ARB (λn. ind_type$BOTTOM)) ∨
                   (∃a.
                      a0' =
                      (λa. ind_type$CONSTR (SUC 0) a (λn. ind_type$BOTTOM)) a) ∨
                   (∃a0 a1.
                      (a0' =
                       (λa0 a1.
                          ind_type$CONSTR (SUC (SUC 0)) ARB
                            (ind_type$FCONS a0
                               (ind_type$FCONS a1 (λn. ind_type$BOTTOM)))) a0 a1) ∧
                      'rexp' a0 ∧ 'rexp' a1) ∨
                   (∃a0 a1.
                      (a0' =
                       (λa0 a1.
                          ind_type$CONSTR (SUC (SUC (SUC 0))) ARB
                            (ind_type$FCONS a0
                               (ind_type$FCONS a1 (λn. ind_type$BOTTOM)))) a0 a1) ∧
                      'rexp' a0 ∧ 'rexp' a1) ∨
                   (∃a.
                      (a0' =
                       (λa.
                          ind_type$CONSTR (SUC (SUC (SUC (SUC 0)))) ARB
                            (ind_type$FCONS a (λn. ind_type$BOTTOM))) a) ∧
                      'rexp' a) ⇒
                   'rexp' a0') ⇒
                'rexp' a0') r ⇔ (dest_rexp (mk_rexp r) = r)
   
   [rexp_size_def]  Definition
      
      |- (∀f f1. rexp_size f f1 EmptyLang = 0) ∧
         (∀f f1 a. rexp_size f f1 (Atom a) = 1 + symbol_size f1 f a) ∧
         (∀f f1 a0 a1.
            rexp_size f f1 (Union a0 a1) =
            1 + (rexp_size f f1 a0 + rexp_size f f1 a1)) ∧
         (∀f f1 a0 a1.
            rexp_size f f1 (Conc a0 a1) =
            1 + (rexp_size f f1 a0 + rexp_size f f1 a1)) ∧
         ∀f f1 a. rexp_size f f1 (Star a) = 1 + rexp_size f f1 a
   
   [star]  Definition
      
      |- star =
         (λA a0.
            ∀star'.
              (∀a0.
                 (a0 = []) ∨ a0 ∈ A ∨ (∃s1 s2. (a0 = s1 ++ s2) ∧ s1 ∈ A ∧ star' s2) ⇒
                 star' a0) ⇒
              star' a0)
   
   [starn_curried_def]  Definition
      
      |- ∀x x1. starn x x1 = starn_tupled (x,x1)
   
   [starn_tupled_primitive_def]  Definition
      
      |- starn_tupled =
         WFREC (@R. WF R ∧ ∀v2 l. R (l,SUC v2 − 1) (l,SUC v2))
           (λstarn_tupled a.
              case a of
                 (l,0) -> I {[]}
              || (l,SUC v3) ->
                   I
                     {s |
                      ∃u v. u ∈ l ∧ v ∈ starn_tupled (l,SUC v3 − 1) ∧ (s = u ++ v)})
   
   [symToStr_def]  Definition
      
      |- (∀s. symToStr (TS s) = s) ∧ ∀s. symToStr (NTS s) = s
   
   [symbol_TY_DEF]  Definition
      
      |- ∃rep.
           TYPE_DEFINITION
             (λa0.
                ∀'symbol' .
                  (∀a0.
                     (∃a.
                        a0 =
                        (λa. ind_type$CONSTR 0 (a,ARB) (λn. ind_type$BOTTOM)) a) ∨
                     (∃a.
                        a0 =
                        (λa. ind_type$CONSTR (SUC 0) (ARB,a) (λn. ind_type$BOTTOM))
                          a) ⇒
                     'symbol' a0) ⇒
                  'symbol' a0) rep
   
   [symbol_case_def]  Definition
      
      |- (∀f f1 a. symbol_case f f1 (NTS a) = f a) ∧
         ∀f f1 a. symbol_case f f1 (TS a) = f1 a
   
   [symbol_repfns]  Definition
      
      |- (∀a. mk_symbol (dest_symbol a) = a) ∧
         ∀r.
           (λa0.
              ∀'symbol' .
                (∀a0.
                   (∃a.
                      a0 = (λa. ind_type$CONSTR 0 (a,ARB) (λn. ind_type$BOTTOM)) a) ∨
                   (∃a.
                      a0 =
                      (λa. ind_type$CONSTR (SUC 0) (ARB,a) (λn. ind_type$BOTTOM))
                        a) ⇒
                   'symbol' a0) ⇒
                'symbol' a0) r ⇔ (dest_symbol (mk_symbol r) = r)
   
   [symbol_size_def]  Definition
      
      |- (∀f f1 a. symbol_size f f1 (NTS a) = 1 + f a) ∧
         ∀f f1 a. symbol_size f f1 (TS a) = 1 + f1 a
   
   [tmnlSym_def]  Definition
      
      |- ∀tmnl. tmnlSym (TS tmnl) = tmnl
   
   [toTmnlSym_def]  Definition
      
      |- ∀s. toTmnlSym s = TS s
   
   [ts2str_def]  Definition
      
      |- ∀s. ts2str (TS s) = s
   
   [union_def]  Definition
      
      |- ∀as bs. union as bs = {s | s ∈ as ∨ s ∈ bs}
   
   [datatype_rexp]  Theorem
      
      |- DATATYPE (rexp EmptyLang Atom Union Conc Star)
   
   [datatype_symbol]  Theorem
      
      |- DATATYPE (symbol NTS TS)
   
   [ldNtsApp]  Theorem
      
      |- ∀l1 l2. ldNts (l1 ++ l2) = ldNts l1 ++ ldNts l2
   
   [listCompLens]  Theorem
      
      |- ∀t t' s2 N rst.
           (t' ++ s2 = t ++ [NTS N] ++ rst) ⇒
           (t = t') ∨ (∃pfx sfx. (t' = t ++ [NTS N] ++ pfx) ∧ (rst = pfx ++ sfx)) ∨
           ∃pfx sfx. (t = pfx ++ sfx) ∧ (t' = pfx)
   
   [memdistNtsApp]  Theorem
      
      |- ∀l1 l2.
           MEM e (distinctldNts l1) ∨ MEM e (distinctldNts l2) ⇔
           MEM e (distinctldNts (l1 ++ l2))
   
   [memdistNtsApp']  Theorem
      
      |- ∀x y.
           MEM e (distinctldNts [x]) ∨ MEM e (distinctldNts [y]) ⇔
           MEM e (distinctldNts [x ++ y])
   
   [memldNts]  Theorem
      
      |- ∀t. MEM (NTS A) (ldNts t) ⇔ ∃e. MEM e t ∧ MEM (NTS A) e
   
   [rexp_11]  Theorem
      
      |- (∀a a'. (Atom a = Atom a') ⇔ (a = a')) ∧
         (∀a0 a1 a0' a1'. (Union a0 a1 = Union a0' a1') ⇔ (a0 = a0') ∧ (a1 = a1')) ∧
         (∀a0 a1 a0' a1'. (Conc a0 a1 = Conc a0' a1') ⇔ (a0 = a0') ∧ (a1 = a1')) ∧
         ∀a a'. (Star a = Star a') ⇔ (a = a')
   
   [rexp_Axiom]  Theorem
      
      |- ∀f0 f1 f2 f3 f4.
           ∃fn.
             (fn EmptyLang = f0) ∧ (∀a. fn (Atom a) = f1 a) ∧
             (∀a0 a1. fn (Union a0 a1) = f2 a0 a1 (fn a0) (fn a1)) ∧
             (∀a0 a1. fn (Conc a0 a1) = f3 a0 a1 (fn a0) (fn a1)) ∧
             ∀a. fn (Star a) = f4 a (fn a)
   
   [rexp_case_cong]  Theorem
      
      |- ∀M M' v f f1 f2 f3.
           (M = M') ∧ ((M' = EmptyLang) ⇒ (v = v')) ∧
           (∀a. (M' = Atom a) ⇒ (f a = f' a)) ∧
           (∀a0 a1. (M' = Union a0 a1) ⇒ (f1 a0 a1 = f1' a0 a1)) ∧
           (∀a0 a1. (M' = Conc a0 a1) ⇒ (f2 a0 a1 = f2' a0 a1)) ∧
           (∀a. (M' = Star a) ⇒ (f3 a = f3' a)) ⇒
           (rexp_case v f f1 f2 f3 M = rexp_case v' f' f1' f2' f3' M')
   
   [rexp_distinct]  Theorem
      
      |- (∀a. EmptyLang ≠ Atom a) ∧ (∀a1 a0. EmptyLang ≠ Union a0 a1) ∧
         (∀a1 a0. EmptyLang ≠ Conc a0 a1) ∧ (∀a. EmptyLang ≠ Star a) ∧
         (∀a1 a0 a. Atom a ≠ Union a0 a1) ∧ (∀a1 a0 a. Atom a ≠ Conc a0 a1) ∧
         (∀a' a. Atom a ≠ Star a') ∧ (∀a1' a1 a0' a0. Union a0 a1 ≠ Conc a0' a1') ∧
         (∀a1 a0 a. Union a0 a1 ≠ Star a) ∧ ∀a1 a0 a. Conc a0 a1 ≠ Star a
   
   [rexp_induction]  Theorem
      
      |- ∀P.
           P EmptyLang ∧ (∀s. P (Atom s)) ∧ (∀r r0. P r ∧ P r0 ⇒ P (Union r r0)) ∧
           (∀r r0. P r ∧ P r0 ⇒ P (Conc r r0)) ∧ (∀r. P r ⇒ P (Star r)) ⇒
           ∀r. P r
   
   [rexp_nchotomy]  Theorem
      
      |- ∀rr.
           (rr = EmptyLang) ∨ (∃s. rr = Atom s) ∨ (∃r r0. rr = Union r r0) ∨
           (∃r r0. rr = Conc r r0) ∨ ∃r. rr = Star r
   
   [rightnt]  Theorem
      
      |- ∀s.
           ¬EVERY isTmnlSym s ⇒
           ∃l1 n l2. (s = l1 ++ [NTS n] ++ l2) ∧ EVERY isTmnlSym l2
   
   [star_cases]  Theorem
      
      |- ∀A a0.
           star A a0 ⇔
           (a0 = []) ∨ a0 ∈ A ∨ ∃s1 s2. (a0 = s1 ++ s2) ∧ s1 ∈ A ∧ star A s2
   
   [star_ind]  Theorem
      
      |- ∀A star'.
           star' [] ∧ (∀s. s ∈ A ⇒ star' s) ∧
           (∀s1 s2. s1 ∈ A ∧ star' s2 ⇒ star' (s1 ++ s2)) ⇒
           ∀a0. star A a0 ⇒ star' a0
   
   [star_rules]  Theorem
      
      |- ∀A.
           star A [] ∧ (∀s. s ∈ A ⇒ star A s) ∧
           ∀s1 s2. s1 ∈ A ∧ star A s2 ⇒ star A (s1 ++ s2)
   
   [starn_def]  Theorem
      
      |- (∀l. starn l 0 = {[]}) ∧
         ∀v2 l.
           starn l (SUC v2) =
           {s | ∃u v. u ∈ l ∧ v ∈ starn l (SUC v2 − 1) ∧ (s = u ++ v)}
   
   [starn_ind]  Theorem
      
      |- ∀P. (∀l. P l 0) ∧ (∀l v2. P l (SUC v2 − 1) ⇒ P l (SUC v2)) ⇒ ∀v v1. P v v1
   
   [symListDiv]  Theorem
      
      |- (pfx ++ sfx = s1 ++ [NTS lhs] ++ s2) ∧ EVERY isTmnlSym pfx ∧
         EVERY isNonTmnlSym sfx ∧ EVERY isTmnlSym s1 ⇒
         (pfx = s1) ∧ (sfx = [NTS lhs] ++ s2)
   
   [sym_r1a]  Theorem
      
      |- isTmnlSym e ⇒ ¬isNonTmnlSym e
   
   [sym_r1b]  Theorem
      
      |- ¬isTmnlSym e ⇔ isNonTmnlSym e
   
   [sym_r2]  Theorem
      
      |- isTmnlSym e ⇒ ¬∃n. e = NTS n
   
   [sym_r3]  Theorem
      
      |- ∀v. EVERY isTmnlSym v ⇒ ¬∃n s1 s2. (v = s1 ++ [n] ++ s2) ∧ isNonTmnlSym n
   
   [sym_r3b]  Theorem
      
      |- ∀v. EVERY isNonTmnlSym v ⇒ ¬∃n s1 s2. (v = s1 ++ [n] ++ s2) ∧ isTmnlSym n
   
   [sym_r4]  Theorem
      
      |- ∀v. EVERY isTmnlSym v ⇒ ¬∃n. MEM n v ∧ ¬isTmnlSym n
   
   [sym_r4b]  Theorem
      
      |- ∀v. ¬EVERY isTmnlSym v ⇒ ∃n. MEM n v ∧ isNonTmnlSym n
   
   [sym_r5]  Theorem
      
      |- ∀v. ¬(∃n s1 s2. (v = s1 ++ [n] ++ s2) ∧ isNonTmnlSym n) ⇒ EVERY isTmnlSym v
   
   [sym_r6]  Theorem
      
      |- ∀v. EVERY isTmnlSym v ⇔ ¬∃n s1 s2. (v = s1 ++ [n] ++ s2) ∧ isNonTmnlSym n
   
   [sym_r7]  Theorem
      
      |- ∀v. ¬EVERY isTmnlSym v ⇒ ∃n s1 s2. (v = s1 ++ [n] ++ s2) ∧ isNonTmnlSym n
   
   [sym_r7b]  Theorem
      
      |- ∀v. ¬EVERY isNonTmnlSym v ⇒ ∃n s1 s2. (v = s1 ++ [n] ++ s2) ∧ isTmnlSym n
   
   [symbol_11]  Theorem
      
      |- (∀a a'. (NTS a = NTS a') ⇔ (a = a')) ∧ ∀a a'. (TS a = TS a') ⇔ (a = a')
   
   [symbol_Axiom]  Theorem
      
      |- ∀f0 f1. ∃fn. (∀a. fn (NTS a) = f0 a) ∧ ∀a. fn (TS a) = f1 a
   
   [symbol_case_cong]  Theorem
      
      |- ∀M M' f f1.
           (M = M') ∧ (∀a. (M' = NTS a) ⇒ (f a = f' a)) ∧
           (∀a. (M' = TS a) ⇒ (f1 a = f1' a)) ⇒
           (symbol_case f f1 M = symbol_case f' f1' M')
   
   [symbol_distinct]  Theorem
      
      |- ∀a' a. NTS a ≠ TS a'
   
   [symbol_induction]  Theorem
      
      |- ∀P. (∀n. P (NTS n)) ∧ (∀t. P (TS t)) ⇒ ∀s. P s
   
   [symbol_nchotomy]  Theorem
      
      |- ∀ss. (∃n. ss = NTS n) ∨ ∃t. ss = TS t
   
   [symlistBdwn]  Theorem
      
      |- (h ++ sfx ++ sfx' = e0 ++ sfx ++ v' ++ [NTS B] ++ w ++ s2) ∧
         EVERY isTmnlSym (h ++ sfx ++ e0 ++ v') ⇒
         ∃s0 s1.
           (sfx' = s0 ++ [NTS B] ++ s1) ∧ EVERY isTmnlSym s0 ∧
           (h ++ sfx ++ s0 = e0 ++ sfx ++ v') ∧ ([NTS B] ++ s1 = [NTS B] ++ w ++ s2)
   
   [symlistdiv3]  Theorem
      
      |- (s1 ++ [NTS lhs] ++ s2 = x ++ m ++ y) ∧ m ≠ [] ∧ EVERY isTmnlSym s1 ∧
         EVERY isTmnlSym x ∧ EVERY isNonTmnlSym m ⇒
         (s1 = x) ∧ ([NTS lhs] ++ s2 = m ++ y)
   
   [symlistnil]  Theorem
      
      |- EVERY isTmnlSym l ∧ EVERY isNonTmnlSym l ⇔ (l = [])
   
   
*)
end
