signature whileLemmasTheory =
sig
  type thm = Thm.thm
  
  (*  Definitions  *)
    val mwhile_def : thm
    val owhile_def : thm
    val terminates_def : thm
  
  (*  Theorems  *)
    val mwhileEndCond : thm
    val mwhileEndState : thm
    val mwhile_COND : thm
    val nthStateProp : thm
    val obua_induct2 : thm
    val owhileEndCond : thm
    val owhileEndState : thm
    val owhile_EQ_NONE : thm
    val owhile_EQ_SOME : thm
    val owhile_thm : thm
  
  val whileLemmas_grammars : type_grammar.grammar * term_grammar.grammar
  
  
(*
   [boolLemmas] Parent theory of "whileLemmas"
   
   [mwhile_def]  Definition
      
      |- ∀g f s.
           mwhile g f s =
           owhile (λopt. case opt of NONE -> F || SOME s -> g s)
             (λopt. case opt of NONE -> NONE || SOME s -> f s) (SOME s)
   
   [owhile_def]  Definition
      
      |- ∀C f s.
           owhile C f s =
           if ∃n. ¬C (FUNPOW f n s) then
             SOME (FUNPOW f (LEAST n. ¬C (FUNPOW f n s)) s)
           else
             NONE
   
   [terminates_def]  Definition
      
      |- ∀C f s. terminates (C,f,s) ⇔ ∃n. ¬C (FUNPOW f n s)
   
   [mwhileEndCond]  Theorem
      
      |- (mwhile C f s = SOME (SOME s')) ⇒ ¬C s'
   
   [mwhileEndState]  Theorem
      
      |- (mwhile C f s = SOME (SOME s')) ∧
         (∀x x'. P x ∧ C x ∧ (f x = SOME x') ⇒ P x') ∧ P s ⇒
         P s'
   
   [mwhile_COND]  Theorem
      
      |- mwhile g f s =
         if g s then
           case f s of NONE -> SOME NONE || SOME s' -> mwhile g f s'
         else
           SOME (SOME s)
   
   [nthStateProp]  Theorem
      
      |- (∀x. P x ∧ C x ⇒ P (f x)) ⇒
         ∀n s. P s ∧ (∀m. m < n ⇒ C (FUNPOW f m s)) ⇒ P (FUNPOW f n s)
   
   [obua_induct2]  Theorem
      
      |- terminates (C,f,s) ∧ (∀x. (C x ⇒ P (f x)) ⇒ P x) ⇒ P s
   
   [owhileEndCond]  Theorem
      
      |- ∀s'. (owhile C f s = SOME s') ⇒ ¬C s'
   
   [owhileEndState]  Theorem
      
      |- (owhile C f s = SOME s') ∧ (∀x. P x ∧ C x ⇒ P (f x)) ∧ P s ⇒ P s'
   
   [owhile_EQ_NONE]  Theorem
      
      |- (owhile C f s = NONE) ⇔ ¬terminates (C,f,s)
   
   [owhile_EQ_SOME]  Theorem
      
      |- (∃s'. owhile C f s = SOME s') ⇔ terminates (C,f,s)
   
   [owhile_thm]  Theorem
      
      |- owhile C f s = if C s then owhile C f (f s) else SOME s
   
   
*)
end
