signature pdaEqCfgTheory =
sig
  type thm = Thm.thm
  
  (*  Definitions  *)
    val adj_curried_def : thm
    val adj_tupled_primitive_def : thm
    val grammar2pda_def : thm
    val isGnf_def : thm
    val nextToStk_def : thm
    val ntslCond_def : thm
    val p2gtrans_def : thm
    val pda2grammar_def : thm
    val toRhs_def : thm
    val transFrmState_primitive_def : thm
    val transSym_primitive_def : thm
    val transToState_primitive_def : thm
    val trans_def : thm
    val tupfrmst_def : thm
    val tupinp_def : thm
    val tupstk_def : thm
    val tuptost_def : thm
    val validGnfProd_def : thm
  
  (*  Theorems  *)
    val adj_def : thm
    val adj_ind : thm
    val everyIsNttoRhs : thm
    val g2pPdaExists : thm
    val gnfNotLastTmns : thm
    val idInpLen : thm
    val idStateEq : thm
    val idcImpEvTmnl : thm
    val idcImpLd : thm
    val idcInpLen : thm
    val idlStateEq : thm
    val idlStkEq : thm
    val isAdj : thm
    val isAdjList : thm
    val isNttoRhs : thm
    val laesImpLang : thm
    val langImpLaes : thm
    val ldImpIdc : thm
    val listPairImpIdc : thm
    val maptransSymtupStk : thm
    val memMapToRhsSl : thm
    val memMapToRhsSl' : thm
    val nextMem2RuleNone : thm
    val nextMem2RuleSome : thm
    val nextMemRuleExistsNone : thm
    val nextMemRuleExistsSome : thm
    val nextMemRuleExistsSome1 : thm
    val notNoneTrans : thm
    val p2gEqPda : thm
    val p2gGrExists : thm
    val p2gImpPda : thm
    val p2gtransNone : thm
    val p2gtransSome : thm
    val pdaImpP2g : thm
    val pdaStackSym : thm
    val pdaStartState : thm
    val pdaTransListDerives : thm
    val pdaTransSplit : thm
    val ruleMem2nextMemNone : thm
    val ruleMem2nextMemSome : thm
    val thm5_3 : thm
    val toRhsLd : thm
    val transFrmStEq : thm
    val transFrmState_def : thm
    val transFrmState_ind : thm
    val transMemList : thm
    val transMemRule : thm
    val transStateEq : thm
    val transSym_def : thm
    val transSym_ind : thm
    val transSymtupStk : thm
    val transToStEq : thm
    val transToState_def : thm
    val transToState_ind : thm
  
  val pdaEqCfg_grammars : type_grammar.grammar * term_grammar.grammar
  
  
(*
   [grammarDef] Parent theory of "pdaEqCfg"
   
   [pdaDef] Parent theory of "pdaEqCfg"
   
   [adj_curried_def]  Definition
      
      |- ∀x x1. adj x x1 ⇔ adj_tupled (x,x1)
   
   [adj_tupled_primitive_def]  Definition
      
      |- adj_tupled =
         WFREC (@R. WF R)
           (λadj_tupled a.
              case a of
                 (NTS (q,e,q'),NTS (r,e',r')) -> I (q' = r)
              || (NTS (q,e,q'),TS v10) -> ARB
              || (TS v3,v1) -> ARB)
   
   [grammar2pda_def]  Definition
      
      |- ∀g q.
           grammar2pda g q =
           (let ts = MAP (trans q) (rules g) in
              <|start := q; ssSym := NTS (startSym g); next := ts; final := []|>)
   
   [isGnf_def]  Definition
      
      |- ∀g. isGnf g ⇔ EVERY validGnfProd (rules g)
   
   [nextToStk_def]  Definition
      
      |- ∀a b c d e. nextToStk ((a,b,c),d,e) = e
   
   [ntslCond_def]  Definition
      
      |- ∀m q' ql mrhs ntsl.
           ntslCond m (q',ql,mrhs) ntsl ⇔
           EVERY isNonTmnlSym ntsl ∧
           (∀e1 e2 p s. (ntsl = p ++ [e1; e2] ++ s) ⇒ adj e1 e2) ∧
           (transFrmState (HD ntsl) = q') ∧ (transToState (LAST ntsl) = ql) ∧
           (∀e. MEM e ntsl ⇒ MEM (transToState e) (statesList m m.next)) ∧
           ∀e. MEM e ntsl ⇒ MEM (transFrmState e) (statesList m m.next)
   
   [p2gtrans_def]  Definition
      
      |- ∀m l ntsl.
           p2gtrans m (rule l ntsl) ⇔
           ∃isymo ssym q q' p mrhs.
             MEM q (statesList m m.next) ∧ MEM q' (statesList m m.next) ∧
             MEM p (statesList m m.next) ∧
             (MEM ((isymo,ssym,q),q',mrhs) m.next ∧ (l = (q,ssym,p))) ∧
             ((ntsl = []) ∧ (isymo = NONE) ∧ (q' = p) ∧ (mrhs = []) ∨
              (∃ts.
                 (ntsl = [TS ts]) ∧ (isymo = SOME (TS ts)) ∧ (q' = p) ∧
                 (mrhs = [])) ∨
              ∃h t.
                ((ntsl = h::t) ∧ t ≠ []) ∧
                ((∃ts.
                    (h = TS ts) ∧ (isymo = SOME (TS ts)) ∧ (MAP transSym t = mrhs) ∧
                    ntslCond m (q',p,mrhs) t) ∨
                 (isymo = NONE) ∧ (MAP transSym ntsl = mrhs) ∧
                 ntslCond m (q',p,mrhs) ntsl))
   
   [pda2grammar_def]  Definition
      
      |- ∀m g.
           pda2grammar m g ⇔
           (∀q.
              MEM q (statesList m m.next) ⇔
              MEM (rule (startSym g) [NTS (m.start,m.ssSym,q)]) (rules g)) ∧
           ∀r. MEM r (rules g) ⇔ p2gtrans m r
   
   [toRhs_def]  Definition
      
      |- ∀frm inp stk tost. toRhs (frm,inp,stk,tost) = NTS (frm,stk,tost)
   
   [transFrmState_primitive_def]  Definition
      
      |- transFrmState =
         WFREC (@R. WF R)
           (λtransFrmState a. case a of NTS (st,sym,st') -> I st || TS v1 -> ARB)
   
   [transSym_primitive_def]  Definition
      
      |- transSym =
         WFREC (@R. WF R)
           (λtransSym a. case a of NTS (st,sym,st') -> I sym || TS v1 -> ARB)
   
   [transToState_primitive_def]  Definition
      
      |- transToState =
         WFREC (@R. WF R)
           (λtransToState a. case a of NTS (st,sym,st') -> I st' || TS v1 -> ARB)
   
   [trans_def]  Definition
      
      |- ∀q l r. trans q (rule l r) = ((SOME (HD r),NTS l,q),q,TL r)
   
   [tupfrmst_def]  Definition
      
      |- ∀st syms stk st'. tupfrmst (st,syms,stk,st') = st
   
   [tupinp_def]  Definition
      
      |- ∀st syms stk st'. tupinp (st,syms,stk,st') = syms
   
   [tupstk_def]  Definition
      
      |- ∀st syms stk st'. tupstk (st,syms,stk,st') = stk
   
   [tuptost_def]  Definition
      
      |- ∀st syms stk st'. tuptost (st,syms,stk,st') = st'
   
   [validGnfProd_def]  Definition
      
      |- ∀l r.
           validGnfProd (rule l r) ⇔
           ∃ts ntsl. (r = ts::ntsl) ∧ isTmnlSym ts ∧ EVERY isNonTmnlSym ntsl
   
   [adj_def]  Theorem
      
      |- adj (NTS (q,e,q')) (NTS (r,e',r')) ⇔ (q' = r)
   
   [adj_ind]  Theorem
      
      |- ∀P.
           (∀q e q' r e' r'. P (NTS (q,e,q')) (NTS (r,e',r'))) ∧
           (∀q e q' v11. P (NTS (q,e,q')) (TS v11)) ∧ (∀v4 v16. P (TS v4) v16) ⇒
           ∀v v1. P v v1
   
   [everyIsNttoRhs]  Theorem
      
      |- ∀l. EVERY isNonTmnlSym (MAP toRhs l)
   
   [g2pPdaExists]  Theorem
      
      |- ∀g. ∃m. m = grammar2pda g
   
   [gnfNotLastTmns]  Theorem
      
      |- ∀x y.
           (lderives g)^* x y ⇒
           (x = [NTS n]) ⇒
           isGnf g ⇒
           ∀p s nts. (y = p ++ [NTS nts] ++ s) ⇒ EVERY isNonTmnlSym s
   
   [idInpLen]  Theorem
      
      |- grammar2pda g q ⊢ (q,inp,stk) → (q',inp',stk') ⇒ LENGTH inp' < LENGTH inp
   
   [idStateEq]  Theorem
      
      |- grammar2pda g q ⊢ (q,inp,stk) → (q',inp',stk') ⇒ (q = q')
   
   [idcImpEvTmnl]  Theorem
      
      |- ∀dl sfx stk stk'.
           ID (grammar2pda g q) ⊢ dl ◁ (q,pfx ++ sfx,stk) → (q,sfx,stk') ∧ isGnf g ⇒
           EVERY isTmnlSym pfx
   
   [idcImpLd]  Theorem
      
      |- ∀dl sfx stk stk'.
           ID (grammar2pda g q) ⊢ dl ◁ (q,pfx ++ sfx,stk) → (q,sfx,stk') ∧ isGnf g ⇒
           (lderives g)^* stk (pfx ++ stk')
   
   [idcInpLen]  Theorem
      
      |- ∀dl q inp inp' stk stk'.
           ID (grammar2pda g q) ⊢ dl ◁ (q,inp,stk) → (q,inp',stk') ∧ LENGTH dl > 1 ⇒
           LENGTH inp' < LENGTH inp
   
   [idlStateEq]  Theorem
      
      |- ∀dl q q' inp inp' stk stk'.
           ID (grammar2pda g q) ⊢ dl ◁ (q,inp,stk) → (q',inp',stk') ⇒ (q = q')
   
   [idlStkEq]  Theorem
      
      |- ∀dl q inp stk stk'.
           ID (grammar2pda g q) ⊢ dl ◁ (q,inp,stk) → (q,inp,stk') ⇒ (stk = stk')
   
   [isAdj]  Theorem
      
      |- ∀e1 e2 pfx sfx.
           (l = pfx ++ [e1; e2] ++ sfx) ⇒
           (tupfrmst e2 = tuptost e1) ⇒
           adj (toRhs e1) (toRhs e2)
   
   [isAdjList]  Theorem
      
      |- ∀t p s.
           (∀e1 e2 pfx sfx.
              (t = pfx ++ [e1; e2] ++ sfx) ⇒ (tupfrmst e2 = tuptost e1)) ⇒
           (MAP toRhs t = p ++ [e1; e2] ++ s) ⇒
           adj e1 e2
   
   [isNttoRhs]  Theorem
      
      |- isNonTmnlSym (toRhs h)
   
   [laesImpLang]  Theorem
      
      |- x ∈ laes (grammar2pda g q) ∧ isGnf g ⇒ x ∈ language g
   
   [langImpLaes]  Theorem
      
      [oracles: DISK_THM, MK_THM, tactic_failed] [axioms: ] []
      |- x ∈ language g ∧ isGnf g ⇒ x ∈ laes (grammar2pda g q)
   
   [ldImpIdc]  Theorem
      
      [oracles: tactic_failed] [axioms: ] []
      |- ∀dl x y.
           lderives g ⊢ dl ◁ [NTS (startSym g)] → (x ++ y) ∧ isGnf g ∧
           EVERY isTmnlSym x ∧ EVERY isNonTmnlSym y ⇒
           (ID (grammar2pda g q))^* (q,x,[NTS (startSym g)]) (q,[],y)
   
   [listPairImpIdc]  Theorem
      
      |- ∀l.
           (∀e1 e2.
              MEM (e1,e2) l ⇒
              (ID m)^* (transFrmState e1,e2,[transSym e1]) (transToState e1,[],[])) ∧
           (∀e1 e2 p s. (MAP FST l = p ++ [e1; e2] ++ s) ⇒ adj e1 e2) ⇒
           EVERY isNonTmnlSym (MAP FST l) ⇒
           ∀h1 h2 i1 i2 t.
             (l = (h1,h2)::t) ∧ (LAST l = (i1,i2)) ⇒
             (ID m)^*
               (transFrmState h1,h2 ++ FLAT (MAP SND t),
                [transSym h1] ++ MAP transSym (MAP FST t)) (transToState i1,[],[])
   
   [maptransSymtupStk]  Theorem
      
      |- ∀l. MAP transSym (MAP toRhs l) = MAP tupstk l
   
   [memMapToRhsSl]  Theorem
      
      |- ∀t.
           MEM e (MAP toRhs t) ∧
           (∀e. MEM e (MAP tuptost t) ⇒ MEM e (statesList m m.next)) ⇒
           MEM (transToState e) (statesList m m.next)
   
   [memMapToRhsSl']  Theorem
      
      |- ∀t.
           MEM e (MAP toRhs t) ∧
           (∀e. MEM e (MAP tupfrmst t) ⇒ MEM e (statesList m m.next)) ⇒
           MEM (transFrmState e) (statesList m m.next)
   
   [nextMem2RuleNone]  Theorem
      
      |- MEM ((NONE,A,q),p,[]) m.next ⇒
         pda2grammar m g ⇒
         MEM (rule (q,A,p) []) (rules g)
   
   [nextMem2RuleSome]  Theorem
      
      |- MEM ((SOME ih,A,q),p,[]) m.next ⇒
         pda2grammar m g ⇒
         MEM (rule (q,A,p) [ih]) (rules g)
   
   [nextMemRuleExistsNone]  Theorem
      
      |- MEM ((NONE,A,q),q',MAP tupstk l) m.next ∧ pda2grammar m g ⇒
         ∃p mrhs. MEM (rule (q,A,p) mrhs) (rules g) ∧ (mrhs = MAP toRhs l)
   
   [nextMemRuleExistsSome]  Theorem
      
      |- MEM ((SOME ih,A,q),q',MAP tupstk l) m.next ∧ pda2grammar m g ∧
         LENGTH l > 1 ⇒
         ∃p ntsl.
           MEM (rule (q,A,p) ntsl) (rules g) ∧
           ∃h t.
             (ntsl = h::t) ∧ t ≠ [] ∧ (h = ih) ∧ (MAP transSym t = MAP tupstk l) ∧
             ntslCond m (q',p,MAP tupstk l) t
   
   [nextMemRuleExistsSome1]  Theorem
      
      |- MEM ((SOME ih,A,q),q',[tupstk h]) m.next ∧ pda2grammar m g ⇒
         MEM (rule (q,A,q') [ih]) (rules g) ∧
         ∃ntsl ts. (ntsl = [TS ts]) ∧ (isymo = SOME (TS ts)) ∧ (mrhs = [TS ts])
   
   [notNoneTrans]  Theorem
      
      |- ∀l q q'. ¬∃sym stk. trans q l = ((NONE,sym,q),q',stk)
   
   [p2gEqPda]  Theorem
      
      [oracles: DISK_THM, MK_THM, tactic_failed] [axioms: ] []
      |- pda2grammar m g ∧ EVERY isTmnlSym x ⇒
         ∀q A p. (derives g)^* [NTS (q,A,p)] x ⇔ (ID m)^* (q,x,[A]) (p,[],[])
   
   [p2gGrExists]  Theorem
      
      |- ∀m. ∃g. g = pda2grammar m
   
   [p2gImpPda]  Theorem
      
      [oracles: DISK_THM, tactic_failed] [axioms: ] []
      |- ∀q A p x.
           derives g ⊢ dl ◁ [NTS (q,A,p)] → x ⇒
           EVERY isTmnlSym x ⇒
           pda2grammar m g ⇒
           (ID m)^* (q,x,[A]) (p,[],[])
   
   [p2gtransNone]  Theorem
      
      |- LENGTH l ≥ 2 ∧ MEM ((NONE,A,q),q',MAP tupstk l) m.next ∧
         (∀e. MEM e (MAP tuptost l) ⇒ MEM e (statesList m m.next)) ∧
         (∀e. MEM e (MAP tupfrmst l) ⇒ MEM e (statesList m m.next)) ∧
         (∀h t. (l = h::t) ⇒ (tupfrmst h = q') ∧ (tuptost (LAST (h::t)) = p)) ∧
         (∀e1 e2 pfx sfx.
            (l = pfx ++ [e1; e2] ++ sfx) ⇒ (tupfrmst e2 = tuptost e1)) ⇒
         p2gtrans m (rule (q,A,p) (MAP toRhs l))
   
   [p2gtransSome]  Theorem
      
      |- LENGTH l ≥ 2 ∧ MEM ((SOME ih,A,q),q',MAP tupstk l) m.next ∧ isTmnlSym ih ∧
         (∀e. MEM e (MAP tuptost l) ⇒ MEM e (statesList m m.next)) ∧
         (∀e. MEM e (MAP tupfrmst l) ⇒ MEM e (statesList m m.next)) ∧
         (∀h t. (l = h::t) ⇒ (tupfrmst h = q') ∧ (tuptost (LAST (h::t)) = p)) ∧
         (∀e1 e2 pfx sfx.
            (l = pfx ++ [e1; e2] ++ sfx) ⇒ (tupfrmst e2 = tuptost e1)) ⇒
         p2gtrans m (rule (q,A,p) (ih::MAP toRhs l))
   
   [pdaImpP2g]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] []
      |- ∀x q A p.
           ID m ⊢ dl ◁ (q,x,[A]) → (p,[],[]) ∧ EVERY isTmnlSym x ∧ pda2grammar m g ⇒
           (derives g)^* [NTS (q,A,p)] x
   
   [pdaStackSym]  Theorem
      
      |- (grammar2pda g q).ssSym = NTS (startSym g)
   
   [pdaStartState]  Theorem
      
      |- (grammar2pda g q).start = q
   
   [pdaTransListDerives]  Theorem
      
      |- ∀l.
           (∀e1 e2 pfx sfx.
              (l = pfx ++ [e1; e2] ++ sfx) ⇒ (tupfrmst e2 = tuptost e1)) ∧
           (∀e.
              MEM e l ⇒
              (derives g)^* [NTS (tupfrmst e,tupstk e,tuptost e)] (tupinp e)) ⇒
           (derives g)^* (MAP toRhs l) (FLAT (MAP tupinp l))
   
   [pdaTransSplit]  Theorem
      
      [oracles: DISK_THM, MK_THM] [axioms: ] []
      |- ∀q qf inp stk.
           ID p ⊢ dl ◁ (q,inp,stk) → (qf,[],[]) ⇒
           ∃l.
             (inp = FLAT (MAP tupinp l)) ∧ (stk = MAP tupstk l) ∧
             (∀e. MEM e (MAP tuptost l) ⇒ MEM e (statesList p p.next)) ∧
             (∀e. MEM e (MAP tupfrmst l) ⇒ MEM e (statesList p p.next)) ∧
             (∀h t.
                (l = h::t) ⇒
                (tupfrmst h = q) ∧ (tupstk h = HD stk) ∧ (tuptost (LAST l) = qf)) ∧
             ∀e1 e2 pfx sfx.
               (l = pfx ++ [e1; e2] ++ sfx) ⇒
               (tupfrmst e2 = tuptost e1) ∧
               ∀e.
                 MEM e l ⇒
                 ∃m.
                   m < LENGTH dl ∧
                   NRC (ID p) m (tupfrmst e,tupinp e,[tupstk e]) (tuptost e,[],[])
   
   [ruleMem2nextMemNone]  Theorem
      
      |- MEM (rule (q,A,p) []) (rules g) ⇒
         pda2grammar m g ⇒
         MEM ((NONE,A,q),p,[]) m.next
   
   [ruleMem2nextMemSome]  Theorem
      
      |- MEM (rule (q,A,p) [ih]) (rules g) ⇒
         pda2grammar m g ⇒
         MEM ((SOME ih,A,q),p,[]) m.next
   
   [thm5_3]  Theorem
      
      [oracles: DISK_THM, MK_THM, tactic_failed] [axioms: ] []
      |- ∀g. isGnf g ⇒ ∃m. x ∈ language g ⇔ x ∈ laes m
   
   [toRhsLd]  Theorem
      
      |- ∀l.
           (∀e.
              MEM e l ⇒
              (derives g)^* [NTS (tupfrmst e,tupstk e,tuptost e)]
                (MAP toTmnlSym (tupinp e))) ⇒
           (derives g)^* (MAP toRhs l) (MAP toTmnlSym (FLAT (MAP tupinp l)))
   
   [transFrmStEq]  Theorem
      
      |- transFrmState (toRhs h) = tupfrmst h
   
   [transFrmState_def]  Theorem
      
      |- transFrmState (NTS (st,sym,st')) = st
   
   [transFrmState_ind]  Theorem
      
      |- ∀P. (∀st sym st'. P (NTS (st,sym,st'))) ∧ (∀v2. P (TS v2)) ⇒ ∀v. P v
   
   [transMemList]  Theorem
      
      |- ∀l.
           MEM ((SOME h,h',q),q,sl) (MAP (trans q) l) ∧ EVERY validGnfProd l ⇒
           ∃nts.
             (h' = NTS nts) ∧
             ∃ts. (h = TS ts) ∧ EVERY isNonTmnlSym sl ∧ MEM (rule nts (h::sl)) l
   
   [transMemRule]  Theorem
      
      |- MEM ((SOME h,h',q),q,sl) (grammar2pda g q).next ∧ isGnf g ⇒
         ∃nts.
           (h' = NTS nts) ∧
           ∃ts.
             (h = TS ts) ∧ EVERY isNonTmnlSym sl ∧ MEM (rule nts (h::sl)) (rules g)
   
   [transStateEq]  Theorem
      
      |- ∀l q q'.
           MEM ((h,h',q),q',sl') (MAP (trans q) l) ∨
           MEM ((h,h',q'),q,sl') (MAP (trans q) l) ⇒
           (q = q')
   
   [transSym_def]  Theorem
      
      |- transSym (NTS (st,sym,st')) = sym
   
   [transSym_ind]  Theorem
      
      |- ∀P. (∀st sym st'. P (NTS (st,sym,st'))) ∧ (∀v2. P (TS v2)) ⇒ ∀v. P v
   
   [transSymtupStk]  Theorem
      
      |- transSym (toRhs h) = tupstk h
   
   [transToStEq]  Theorem
      
      |- transToState (toRhs h) = tuptost h
   
   [transToState_def]  Theorem
      
      |- transToState (NTS (st,sym,st')) = st'
   
   [transToState_ind]  Theorem
      
      |- ∀P. (∀st sym st'. P (NTS (st,sym,st'))) ∧ (∀v2. P (TS v2)) ⇒ ∀v. P v
   
   
*)
end
